/*
 * Netris -- A free networked version of T*tris
 * Copyright (C) 1994,1995,1996  Mark H. Weaver <mhw@netris.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define NOEXT
#include "netris.h"

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <setjmp.h>

#include "util.h"

#define HEADER_SIZE sizeof(netint4[3])

static struct option options[] = {
	{ "wait",        0, 0, 'w' },
	{ "port",        1, 0, 'p' },
	{ "quadra",      1, 0, 'q' },
	{ "min-players", 1, 0, 'm' },
	{ "max-players", 1, 0, 'x' },
	{ "continuous",  1, 0, 'c' },
	{ "speed",       1, 0, 'i' },
	{ "seed",        1, 0, 's' },
	{ "verbose",     0, 0, 'v' },
	{ "info",        0, 0, 'H' },
	{ "help",        0, 0, 'h' },
	{ 0,             0, 0,  0  }
};

static char minplayers = 2;
static char maxplayers = 8;
static char playercount;
static char verbose = 0;

struct sockaddr_in addr;

MyEventType NetGenFunc(EventGenRec *gen, MyEvent *event);
static EventGenRec netGen[MAX_SCREENS] = {
	{ NULL, 0, FT_read, -1, NetGenFunc, EM_net, 0, "\0", 0, HEADER_SIZE }
};

static MyEventType AlarmGenFunc(EventGenRec *gen, MyEvent *event);
static EventGenRec alarmGen = {
	&alarmGen, 0, FT_read, -1, AlarmGenFunc, EM_alarm
};

static MyEventType ConnGenFunc(EventGenRec *gen, MyEvent *event);
static EventGenRec connGen = {
	NULL, 0, FT_read, -1, ConnGenFunc, EM_connect
};

static EventGenRec *nextGen = &alarmGen;

static sigjmp_buf close_env;


void SendPacketTo(short playa, short uid, NetPacketType type, int size, void *data)
{ //send to someone
	netint4 header[3];

	if (netGen[playa].fd >= 0) {
		if (verbose)
			fprintf(stderr, ": send %d from %d to %d\n", type, uid, playa);
		header[0] = hton4(uid);
		header[1] = hton4(type);
		header[2] = hton4(size + HEADER_SIZE);
		if (MyWrite(netGen[playa].fd, header, HEADER_SIZE) != HEADER_SIZE)
			die("write (header)");
		if (size > 0 && data && MyWrite(netGen[playa].fd, data, size) != size)
			die("write");
	}
}

static MyEventType AlarmGenFunc(EventGenRec *gen, MyEvent *event)
{
	return E_alarm;
}

void SCloseNet(short playa)
{ //kick some connection's ass!
	MyEvent event;

	if (netGen[playa].fd >= 0) {
		if (Players[playa].alive >= 0) {
			SendPacketTo(playa, 0, NP_endConn, 0, NULL);
			do {} while (WaitMyEvent(&event, EM_net) != E_lostConn);
		} //say bye to player
		close(netGen[playa].fd);
		netGen[playa].fd = -1;
	}
	if (netGen[playa].next)
		RemoveEventGen(&netGen[playa]);
}

void CloseNets(void)
{ //nou oogjes dicht en snaveltjes toe
	int i;

	fprintf(stderr, "- Closing connections...\n");
	for (i = 1; i < MAX_SCREENS; i++)
		SCloseNet(i); //bye everybuddy
	fprintf(stderr, "* All Done\n\n");
}

MyEventType NetGenFunc(EventGenRec *gen, MyEvent *event)
{ //receive
	int result;
	short uid, type, size;
	netint4 *data = (netint4*)&(gen->buf);

	result = MyRead(gen->fd, gen->buf + gen->bufSize,
		gen->bufGoal - gen->bufSize);
	if (result < 0) {
		fprintf(stderr, "- Closed connection to player #%d\n", gen->player);
		return E_lostConn;
	}
	gen->bufSize += result;
	if (gen->bufSize < gen->bufGoal)
		return E_none;
	// *ugly* memcpy(data, gen->buf, sizeof(data));
	uid = ntoh4(data[0]);
	type = ntoh4(data[1]);
	size = ntoh4(data[2]);
	gen->bufGoal = size;
	if (gen->bufSize < gen->bufGoal)
		return E_none;
	gen->bufSize = 0;
	gen->bufGoal = HEADER_SIZE;
	event->u.net.sender = gen->player;
	event->u.net.uid = uid;
	event->u.net.type = type;
	event->u.net.size = size - HEADER_SIZE;
	event->u.net.data = gen->buf + HEADER_SIZE;
	if (type == NP_endConn) {
		fprintf(stderr, "- Quit player #%d\n", gen->player);
		return E_lostConn;
	} //client sent quit signal
	return E_net;
}


static MyEventType ConnGenFunc(EventGenRec *gen, MyEvent *event)
{
	int addrLen;
	struct linger val2;
	int new;

	addrLen = sizeof(addr);
	for (new = 1; new <= MAX_SCREENS; new++)
		if (netGen[new].fd < 0) break;
	if (new > maxplayers) return;

	if ((netGen[new].fd =
		accept(gen->fd, (struct sockaddr *)&addr, &addrLen)) < 0)
			die("accept");
	fprintf(stderr, "+ Connection: %s\n", inet_ntoa(addr.sin_addr));
	val2.l_onoff = 1;
	val2.l_linger = 0;
	setsockopt(netGen[new].fd, SOL_SOCKET, SO_LINGER,(void *)&val2,
		sizeof(val2));
	AddEventGen(&netGen[new]);
	netGen[new].player = event->u.net.uid = new;
	{ //new connection; netGen already initialized in GenFunc
		struct hostent *host;

		sprintf(Players[new].host, "%s", inet_ntoa(addr.sin_addr));
		if (addr.sin_family == AF_INET) {
			host = gethostbyaddr((void *)&addr.sin_addr,
								 sizeof(struct in_addr), AF_INET);
			if (host) {
				strncpy(Players[new].host, host->h_name,
					sizeof(Players[new].host) - 1);
				Players[new].host[sizeof(Players[new].host) - 1] = 0;
			} //set player's hostname
		}
	} //E_connect
	return E_connect;
}

void CountPlayers(void)
{ //count number of players/teams
	int i, j;
	playercount = 0;
	for (i = 1; i < MAX_SCREENS; i++) if (Players[i].alive > 0) {
		if (Players[i].team < 128) for (j = 1; j < i; j++)
			if (Players[j].alive > 0 && (Players[j].team == Players[i].team)) {
				playercount--;
				break;
			} //player of same team counted before
		playercount++;
	} //player alive
}

int StartServer(void)
{
	MyEvent event;
	netint2 currentpiece[MAX_SCREENS];
	int playersReady = 0;
	int paused = 1;
	int i;
	char teams[10][7] = {
		"", "Green", "Cyan", "Blue", "Purple",
		"Red", "Grey", "White", "*Orange"
	};

	do {
		switch (WaitMyEvent(&event, EM_any)) {
			case E_lostConn: //client went away :(
				Players[event.u.net.sender].alive = -1;
				for (i = 1; i < MAX_SCREENS; i++)
					if (Players[i].alive >= 0)
						SendPacketTo(i, event.u.net.sender,
							NP_part, sizeof(Players[0].alive),
							&Players[event.u.net.sender].alive);
				SCloseNet(event.u.net.sender);
				break; //NP_endConn
			case E_net:
				if (verbose) fprintf(stderr, ": %d sent %d\n",
					netGen[event.u.net.sender].fd, event.u.net.type);
				switch(event.u.net.type) {
				case NP_hello:
//					if (event.u.net.type != NP_hello) ByeClient(new);
				{
					netint4 versiondata[2];
					char data[255];
					int major;
					int protocolVersion;

					memcpy(versiondata, event.u.net.data,
						sizeof(versiondata));
					major = ntoh4(versiondata[0]);
					protocolVersion = ntoh4(versiondata[1]);
					if (major != MAJOR_VERSION
					 || protocolVersion != PROTOCOL_VERSION) {
						snprintf(data, sizeof(data),
							"Version mismatch: received %d.%d",
							major, protocolVersion);
						fprintf(stderr, "= Wrong version player #%d (%s)\n",
							event.u.net.sender, data);
						SendPacketTo(event.u.net.sender, 0, NP_error,
							strlen(data)+1, data);
						SCloseNet(event.u.net.sender);
					} //version mismatch
					fprintf(stderr, "* Accepted player #%d\n",
						event.u.net.sender);
					break;
				} //NP_hello
				case NP_newPlayer:
				//receive player details and return other players
					memcpy(&Players[event.u.net.sender],
						event.u.net.data, event.u.net.size);
					if (Players[event.u.net.sender].team < 1
					|| Players[event.u.net.sender].team > 7) {
						int team;

						for (team = 1; team < 7; team++) {
							for (i = 1; i < MAX_SCREENS; i++)
								if ((Players[i].alive > 0) && (Players[i].team == team))
									break; //team in use
							if (i==MAX_SCREENS) break;
						} //find unused team
						Players[event.u.net.sender].team = team;
						SendPacketTo(event.u.net.sender, event.u.net.sender, NP_team,
							sizeof(Players[event.u.net.sender].team),
							&Players[event.u.net.sender].team);
					} //invalid team
					if (Game.started < 2)
						Players[event.u.net.sender].flags |= SCF_paused;
					if (!Game.continuous && Game.started >= 2) {
						char data[40];
						strcpy(data, "Can't join: Game has already started");
						fprintf(stderr, "- Can't join player #%d in "
							"non-continuous game\n", event.u.net.sender);
						SendPacketTo(event.u.net.sender, 0, NP_error,
							strlen(data)+1, data);
//						SCloseNet(event.u.net.sender, 0);
						break;
					} //can't join started game
					{
						static struct {
							int playerflags;
							int gravity;	//1
							int started;	//2
							int continuous;	//3
							long seed;		//4
							int initspeed;	//5
						} data;

						memcpy(&data, &Players[event.u.net.sender].flags,
							sizeof(data.playerflags));
						memcpy(&data.gravity, &Game,
							sizeof(data) - sizeof(data.playerflags));
						SendPacketTo(event.u.net.sender, 0, NP_gamedata,
							sizeof(data), &data);
					} //send game options
					for (i = 1; i < MAX_SCREENS; i++)
						if (netGen[i].fd >= 0 && i != event.u.net.sender) {
							SendPacketTo(event.u.net.sender, i,
								NP_newPlayer, sizeof(_Player), &Players[i]);
							SendPacketTo(event.u.net.sender, i, NP_newPiece,
								sizeof(Players[i].curShape), &Players[i].curShape);
							SendPacketTo(i, event.u.net.sender, NP_newPlayer,
								sizeof(_Player), &Players[event.u.net.sender]);
						} //send (to) players
					fprintf(stderr, "> Joined player #%d: %s <%s> (%s)\n",
						event.u.net.sender,
						Players[event.u.net.sender].name,
						Players[event.u.net.sender].host,
						teams[Players[event.u.net.sender].team]);
					if (++playersReady >= minplayers) {
						if (Game.started > 1)
							SendPacketTo(event.u.net.sender, 0,
								NP_start, 0, NULL);
/*						else {
							fprintf(stderr, "* Starting game (%010d)\n",
								Game.seed);
							for (i = 1; i < MAX_SCREENS; i++)
								SendPacketTo(i, 0, NP_start, 0, NULL);
							Game.started++;
						} //first goahead (to all)*/
					} //give go ahead
					break; //NP_playerdata
				case NP_newPiece:
					memcpy(&Players[event.u.net.sender].curShape,
						event.u.net.data, sizeof(Players[0].curShape));
					goto sendtoall;
				case NP_argghhh:
					Players[event.u.net.sender].alive = 0;
					fprintf(stderr, "< Player #%d died\n",
						event.u.net.sender);
					//check for unpaused game
				case NP_pause:
				{
					Players[event.u.net.sender].flags ^= SCF_paused;
					paused = Game.started < 1;
					for (i = 1; i < MAX_SCREENS; i++)
						if (Players[i].alive > 0)
							paused |= Players[i].flags & SCF_paused;
					fprintf(stderr, "* Player #%d (un)paused (pause=%d)\n",
						event.u.net.sender, paused);
					if (paused) paused = 1;
					goto sendtoall;
				} //NP_pause
				default: //relay data to all players
				sendtoall:
//					if (event.u.net.type >= NP_pause)
					if (event.u.net.type >= NP_rotright
					 && Game.started < 2)
						break;
					for (i = 1; i < MAX_SCREENS; i++)
						if (i != event.u.net.sender)
						if (event.u.net.type != NP_giveJunk ||
						Players[i].team != Players[event.u.net.sender].team)
							SendPacketTo(i, event.u.net.sender,
								event.u.net.type, event.u.net.size,
								event.u.net.data);
					break; //>=NP_paused
				}
				break; //E_net
			case E_connect:
			{ //new connection; netGen already initialized in GenFunc
				char serverdata[255];

				sprintf(serverdata, "Netris server %s", version_string);
				SendPacketTo(event.u.net.uid, event.u.net.uid, NP_hello,
					strlen(serverdata)+1, serverdata);
				break;
			} //E_connect
		} //event
		CountPlayers();
		if (Game.started < 1) {
			if (playercount > 1) {
				fprintf(stderr, "* Game (%010d) ready to start\n", Game.seed);
				Game.started++;
			} //give goahead
		} //game not yet started
		else {
			if (playercount < 2) {
				fprintf(stderr, "* Stopping game\n");
				if (Game.seed) Game.seed++;
				if (Game.started > 1) for (i = 1; i < MAX_SCREENS; i++)
					if (Players[i].alive >= 0) {
						Players[i].alive = 1;
						Players[i].flags |= SCF_paused;
						SendPacketTo(i, 0, NP_stop,
							sizeof(Game.seed), &Game.seed);
					} //transmit game stop and set players not ready
				paused = 1;
				Game.started = 0;
			} //too few players for game
			if (Game.started == 1 && !paused) {
				Game.started++;
				fprintf(stderr, "* Game starts\n");
				for (i = 1; i < MAX_SCREENS; i++)
					if (Players[i].alive > 0)
						SendPacketTo(i, 0, NP_start, 0, NULL);
			} //everybody ready to start
		} //game (ready to) start(ed)
	} while (1);
	fprintf(stderr, "* Exiting server\n");
}


void SHeader(void)
{
	fprintf(stderr,
		"NETRIS Server %s\t(c) 2002 Shiar <shiar@shiar.org>\n\n",
		version_string);
}

void SUsage(void)
{
	SHeader();
	fprintf(stderr,
		"Usage: netris <options>\n"
		"\n"
		"  -h, --help\t\tPrint this usage information\n"
		"  -H, --info\t\tShow distribution and warranty information\n"
		"\n"
		"  -p, --port <port>\tSet port number (default is %d)\n"
		"\n"
		"  -s, --seed <seed>\tStart with given random seed\n"
		"  -i, --speed <sec>\tSet the initial step-down interval, in seconds\n"
		"  -m, --min-players <2>\tNumber of players required before starting the game\n"
		"  -x, --max-players <8>\tMaximum number of players allowed in the game\n"
		"  -c, --continuous\tDon'n quit the game\n"
		"\n", DEFAULT_PORT);
}

void WriteConf(void)
{
	FILE *file_out;

	file_out = fopen(CONFIG_FILE, "w");
	if (file_out == NULL) {
		perror("Error writing config file");
		exit(1);
	}

	fprintf(file_out, "### NETRIS %s Config file ###\n\n", version_string);

	fclose(file_out);
	fprintf(stderr, "Wrote new game configuration to %s\n", CONFIG_FILE);
}

void HandleOption(char tag, char *value)
{
	switch (tag) {
	case 'v':  //verbose
		verbose = 1;
		break;
	case 'p':  //port
		port = atoi(value);
		break;
	case 'c':  //min-players
		Game.continuous = atoi(value);
		break;
	case 'm':  //min-players
		minplayers = atoi(value);
		break;
	case 'x':  //max-players
		maxplayers = atoi(value);
		if (maxplayers >= MAX_SCREENS)
			maxplayers = MAX_SCREENS;
		break;
	case 'q':  //quadra-style gravity
		Game.gravity ^= 1;
		break;
	case 'i':  //speed (of level 1)
		Game.initspeed = atof(value) * 1e6;
		break;
	case 's':  //seed
		Game.seed = atoi(value);
		break;
	case 'H':  //info
		SHeader();
		DistInfo(); exit(0);
	case 'h':  //help
		SUsage(); exit(0);
	default:
		break;
	}
}

void ReadConf(char *filename)
{
	FILE *file_in;
	char buf[513];
	int i;
	char *ch;
	char tag[81], value[81];

	file_in = fopen(filename, "r");
	if (file_in) {
		while (fgets(buf, 512, file_in) != NULL) {
			if ((ch = strchr(buf, '#')))
				*ch = '\0'; // truncate string from # char
			for (i = strlen(buf)-1; i >= 0; i--)
				if (buf[i] == ' '  || buf[i] == '\t'
				 || buf[i] == '\n' || buf[i] == 13)
					buf[i] = '\0';
				else break;

			sscanf(buf, "%80[^= \t] = %80[^\n]", tag, value);
			for (i = 0; options[i].name; i++){
				if (!strcasecmp(options[i].name, tag)) {
					HandleOption(options[i].val, value);
					break;
				}
			}
		}
		fclose(file_in);
	} //read file
	else {
		fprintf(stderr, "Unable to open config file %s.\n", filename);
	} //defaults

}

void CatchInt(int sig)
{
	siglongjmp(close_env, 1);
}

int main(int argc, char **argv)
{
	char ch;

	if (sigsetjmp(close_env, 1)) exit(0);
	signal(SIGINT, CatchInt);
	port = DEFAULT_PORT;
	maxplayers = 8;
	Game.initspeed = DEFAULT_INTERVAL;
	Game.seed = time(0);
	Game.gravity = 0;
	{
		int i;

		for (i = 1; i < MAX_SCREENS; i++)
			Players[i].alive = -1;
	}

//	if (getopt(argc, argv, "f:") == 'f')
//		ReadConf(optarg);
//	else
	ReadConf(CONFIG_FILE);
	while ((ch = getopt_long(
		argc, argv, "hHvqp:i:s:c:m:x:", options, NULL
	)) != -1)
		HandleOption(ch, optarg);
	if (optind < argc) {
		SUsage();
		exit(1);
	}
//	WriteConf();

	SHeader();

	{
		int i;

		for (i = 1; i < MAX_SCREENS; i++)
			memcpy(&netGen[i], &netGen[0], sizeof(EventGenRec));
	} //setup netGen var

	AtExit(CloseNets);

	{
		int val1;

		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_port = htons(port);
		addr.sin_addr.s_addr = htonl(INADDR_ANY);
		if ((connGen.fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
			die("socket");
		val1 = 1;
		setsockopt(connGen.fd, SOL_SOCKET, SO_REUSEADDR,
			(void *)&val1, sizeof(val1));
		if (bind(connGen.fd, (struct sockaddr *)&addr, sizeof(addr)) < 0)
			die("bind");
		if (listen(connGen.fd, 1) < 0)
			die("listen");

		AddEventGen(&connGen);
	} //setup listening sock

	StartServer(); //server loop

	return 0;
}

